'''
    File: mappings.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: September 23, 20203
    
    Classes that represent mappings of utilities/grades to candidates.
'''

import functools
import numpy as np
from pref_voting.rankings import Ranking

# turn off future warnings.
# getting the following warning when calling tabulate to display a profile: 
# /Library/Frameworks/Python.framework/Versions/3.8/lib/python3.8/site-packages/tabulate.py:1027: FutureWarning: elementwise comparison failed; returning scalar instead, but in the future will perform elementwise comparison
#  if headers == "keys" and not rows:
# see https://stackoverflow.com/questions/40659212/futurewarning-elementwise-comparison-failed-returning-scalar-but-in-the-futur
#
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)

# some helper functions

def val_map_function(v, val_map):
    return "None" if v is None else val_map[v]

def default_compare_function(v1, v2):
    if v1 > v2:
        return 1
    elif v2 > v1:
        return -1
    else:
        return 0

class _Mapping(object): 
    """
    A partial function on a set of items. 

    Attributes:
        mapping: a dictionary representing the mapping
        domain: the domain of the mapping
        codomain: the codomain of the mapping
        item_map: a dictionary mapping items to their names
        val_map: a dictionary mapping values to their names
        compare_function: a function used to compare values 
    """
    def __init__(
        self, 
        mapping, # a dictionary representing the mapping
        domain=None, # if domain is None, then it is assumed to be all keys (so mapping is a total function)
        codomain=None, # if codomain is None, then it is assumed to be any number
        compare_function=None, # function used to compare values
        item_map=None, # a dictionary mapping items to their names
        val_map=None, # a dictionary mapping values to their names
        ):
        
        assert domain is None or all([x in domain for x in mapping.keys()]), f"Not all keys in {mapping} are in the domain {domain}."
        
        assert domain is not None or all([isinstance(v, (int, float)) for v in mapping.values()]), f"Not all values in {mapping} are numbers."
        
        assert codomain is None or all([v in codomain for v in mapping.values()]), f"Not all values in {mapping} are in the codomain {codomain}."

        self.mapping = mapping
        self.domain = domain if domain is not None else sorted(list(mapping.keys())) # if domain is None, then it is assumed to be all keys (so mapping is a total function)
        self.codomain = codomain # if codomain is None, then it is assumed to be any number
        self.item_map = item_map if item_map is not None else {x:str(x) for x in self.domain} # a dictionary mapping items in the domain to their names

        val_map = val_map if val_map is not None else {v:str(v) for v in self.mapping.values()} # a dictionary mapping values to their names
        self.val_map = functools.partial(val_map_function, val_map = val_map) # a function mapping values to their names
        
        self.compare_function = compare_function if compare_function is not None else default_compare_function

    def val(self, x): 
        """
        The value assigned to x by the mapping. If x is in the domain but not defined by the mapping, then None is returned.
        """
        assert x in self.domain, f"{x} not in the domain {self.domain}"
        return self.mapping[x] if x in self.mapping.keys() else None

    def has_value(self, x):
        return x in self.mapping.keys()

    @property
    def defined_domain(self):
        return sorted(list(self.mapping.keys()))

    def inverse_image(self, v):
        """Return all the elements in the domain that are mapped to v."""
        return [x for x in self.domain if self.val(x) == v]
        
    def image(self, items=None): 
        """
        The image of the mapping.
        """
        items = self.defined_domain if items is None else items
        return list([self.val(x) for x in items if self.val(x) is not None])

    @property
    def range(self): 
        """
        The range of the mapping.
        """
        return sorted(list(set(self.mapping.values())))

    def average(self, **kwargs): 
        """
        Return the average utility of all elements in alternatives. If alternatives is None, then find the average utility of all elements that are assigned a utility.
        """

        items = kwargs.get("items", None) or kwargs.get("candidates", None) or kwargs.get("alternatives", None)
                
        assert items is None or all([isinstance(self.val(x), (int, float)) or self.val(x) is None for x in items]), "Not all values are numbers."

        img = self.image(items=items)
        return np.mean(img) if len(img) > 0 else None

    def median(self, **kwargs): 
        """
        Return the median utility of all elements in alternatives. If alternatives is None, then find the average utility of all elements that are assigned a utility.
        """

        items = kwargs.get("items", None) or kwargs.get("candidates", None) or kwargs.get("alternatives", None)

        assert items is None or all([isinstance(self.val(x), (int, float)) or self.val(x) is None for x in items]), "Not all values are numbers."

        img = self.image(items=items)
        return np.median(img) if len(img) > 0 else None

    def compare(self, x, y): 
        """
        Return 1 if the value of x is greater than the value of y, 0 if they are equal, and -1 if the value of x is less than the value of y.

        If either x or y is not defined, then None is returned.
        """
        assert x in self.domain, f"{x} not in the domain {self.domain}"
        assert y in self.domain, f"{y} not in the domain {self.domain}"

        return None if (not self.has_value(x) or not self.has_value(y)) else self.compare_function(self.val(x), self.val(y))

    def extended_compare(self, x, y): 
        """
        Return 1 if the value of x is greater than the value of y or x has a value and y does not have a value, 0 if they are equal or both do not have values, and -1 if the value of y is greater than the value of x or y has a value and x does not have a value.

        If either x or y is not defined, then None is returned.
        """
        assert x in self.domain, f"{x} not in the domain {self.domain}"
        assert y in self.domain, f"{y} not in the domain {self.domain}"

        if self.has_value(x) and not self.has_value(y): 
            return 1
        elif self.has_value(y) and not self.has_value(x):
            return -1
        elif not self.has_value(x) and not self.has_value(y):
            return 0
        else:
            return self.compare_function(self.val(x), self.val(y))

    def strict_pref(self, x, y):
        """Returns True if ``x`` is strictly preferred to ``y``.

        The return value is True when both ``x`` and ``y`` are assigned values and the value of ``x`` is strictly greater than the utility of ``y`` according to the compare function.
        """
        return self.compare(x, y) == 1

    def extended_strict_pref(self, x, y):
        """Returns True if ``x`` is strictly preferred to ``y`` using the extended compare function.

        The return value is True when the value of ``x`` is strictly greater than the value of ``y`` or ``x`` is assigned a value and ``y`` is not assigned a value.
        """
        return self.extended_compare(x, y) == 1

    def indiff(self, x, y):
        """Returns True if ``x`` is indifferent with ``y``.

        The return value is True when both ``x`` and ``y`` are assigned values and the value of ``x`` equals the value of ``y``.
        """
        return self.compare(x, y) == 0

    def extended_indiff(self, x, y):
        """Returns True if ``x`` is indifferent with ``y`` using the ``extended_compare`` function.

        The return value is True when the value of ``x`` equals the value of ``y`` or both ``x`` and ``y`` are not assigned values.
        """
        return self.extended_compare(x, y) == 0

    def weak_pref(self, x, y):
        """Returns True if ``x`` is weakly preferred to ``y``.

        The return value is True when both ``x`` and ``y`` are assigned utilities and the utility of ``x`` is at least as  greater than the utility of ``y``.
        """

        return self.strict_pref(x, y) or self.indiff(x, y)

    def extended_weak_pref(self, x, y):
        """Returns True if ``x`` is weakly preferred to ``y``.

        The return value is True when both ``x`` and ``y`` are assigned utilities and the utility of ``x`` is at least as  greater than the utility of ``y``.
        """

        return self.extended_strict_pref(x, y) or self.extended_indiff(x, y)

    def _indifference_classes(self, items, use_extended=False):
        """
        Return a list of the indifference classes of the items.
        """
        indiff_classes = list()
        processed_items = set()
        compare_fnc = self.extended_compare if use_extended else self.compare
        for x in items:
            if x not in processed_items:
                indiff = [y for y in items if compare_fnc(x, y) == 0]
                if len(indiff) > 0:
                    indiff_classes.append(indiff)
                for y in indiff:
                    processed_items.add(y)
        return indiff_classes

    def sorted_domain(self, extended=False):
        """
        Return a list of the indifference classes sorted according to the compare function (or extended compare function if extended is True).
        """
        indiff_classes = self._indifference_classes(self.domain, use_extended=True) if extended else self._indifference_classes(self.defined_domain)

        compare_fnc = self.extended_compare if extended else self.compare

        key_func = functools.cmp_to_key(lambda xs, ys : compare_fnc(xs[0], ys[0]))

        return sorted(indiff_classes, key=key_func, reverse=True)

    def as_dict(self):
        """
        Return the mapping as a dictionary.
        """
        return {c: self.val(c) for c in self.defined_domain}

    def display_str(self, func_name): 
        """
        Return a string representation of the mapping.
        """
        return f"{', '.join([f'{func_name}({self.item_map[x]}) = {self.val_map(self.val(x))}' for x in self.domain])}"
        
    def __call__(self, x): 
        return self.val(x)

    def __repr__(self): 
        return f"{self.mapping}"
    
    def __str__(self):
        return f'{", ".join([f"{self.item_map[x]}:{self.val_map(self.val(x))}" for x in self.domain])}'
    
########
# Utility Functions
########

class Utility(_Mapping):
    def __init__(self, utils, **kwargs):
        """Constructor method for the Utility class."""

        if "domain" in kwargs and "candidates" in kwargs:
            raise ValueError("You can only provide either 'domain' or 'candidates', not both.")
        if "domain" in kwargs:
            self.domain = kwargs["domain"]
        elif "candidates" in kwargs:
            self.domain = kwargs["candidates"]
        else:
            self.domain = list(utils.keys())

        self.cmap = {x:str(x) for x in self.domain} if "cmap" not in kwargs else kwargs["cmap"]

        assert self.domain is None or all([x in self.domain for x in utils.keys()]), f"The domain {self.domain} must contain all elements in the utility map {utils}"
        
        super().__init__(utils, domain=self.domain, item_map = self.cmap)

    @property
    def candidates(self):
        return self.domain

    def items_with_util(self, u):
        """Returns a list of the items that are assigned the utility ``u``."""
        return self.inverse_image(u)

    def has_utility(self, x): 
        """Return True if x has a utility."""
        return self.has_value(x)
    
    def remove_cand(self, x):
        """Returns a utility with the item ``x`` removed."""

        new_utils  = {y: self.val(y) for y in self.defined_domain if y != x}
        new_domain = [y for y in self.domain if y != x]
        new_cmap   = {y: self.cmap[y] for y in self.cmap.keys() if y != x}
        return Utility(new_utils, domain=new_domain, cmap=new_cmap)
    
    def to_approval_ballot(self, prob_to_cont_approving=1.0, decay_rate=0.0): 
        """
        Return an approval ballot representation of the mapping.  It is assumed that the voter approves of all candidates with a utility greater than the average of the utilities assigned to the candidates.

        The parameter ``prob_to_cont_approving`` is the probability that the voter continues to approve of candidates after the first candidate is approved.  The parameter ``decay_rate`` is the exponential decay rate constant in the exponential decay of the probability to continue approving.
        """
        avg_grade = self.average()

        main_approval_set = {x:self.val(x) for x in self.defined_domain if self.val(x) > avg_grade}

        sorted_approval_set = sorted(main_approval_set.items(), key=lambda a: a[1], reverse=True)

        # initialize approval set with the candidate with the highest utility
        approval_set = [sorted_approval_set[0][0]]

        t = 0
        for x, u in sorted_approval_set[1:]:
            if np.random.rand() < prob_to_cont_approving * np.exp(-decay_rate * t):
                approval_set.append(x)
                t += 1
            else:
                break
            
        return Grade(
            {
                x: 1 if x in approval_set else 0 for x in self.defined_domain
            },
            [0, 1],
            candidates=self.domain,
            cmap=self.cmap
        )

    def to_k_approval_ballot(self, k, prob_to_cont_approving=1.0, decay_rate=0.0): 
        """
        Return an k-approval ballot representation of the mapping.  It is assumed that the voter approves of the top k candidates with a utility greater than the average of the utilities assigned to the candidates.

        The parameter ``prob_to_cont_approving`` is the probability that the voter continues to approve of candidates after the first candidate is approved.  The parameter ``decay_rate`` is the exponential decay rate constant in the exponential decay of the probability to continue approving.
        """
        avg_grade = self.average()

        main_approval_set = {x:self.val(x) for x in self.defined_domain if self.val(x) > avg_grade}

        sorted_approval_set = sorted(main_approval_set.items(), key=lambda a: a[1], reverse=True)

        # initialize approval set with the candidate with the highest utility
        approval_set = [sorted_approval_set[0][0]]

        t = 0
        for x, u in sorted_approval_set[1:]:
            if np.random.rand() < prob_to_cont_approving * np.exp(-decay_rate * t):
                approval_set.append(x)
                t += 1
            else:
                break

            if len(approval_set) == k:
                break
            
        return Grade(
            {
                x: 1 if x in approval_set else 0 for x in self.defined_domain
            },
            [0, 1],
            candidates=self.domain,
            cmap=self.cmap
        )

    def ranking(self): 
        """Return the ranking generated by this utility function."""
        return Ranking(
            {x:idx + 1 for idx, indiff_class in enumerate(self.sorted_domain()) 
             for x in indiff_class})
    
    def extended_ranking(self): 
        """Return the ranking generated by this utility function."""

        return Ranking(
            {x:idx + 1 for idx, indiff_class in enumerate(self.sorted_domain(extended=True))
             for x in indiff_class})

    def has_tie(self, use_extended=False): 
        """Return True when there are at least two candidates that are assigned the same utility."""
        return any([len(cs) != 1 for cs in self.sorted_domain(extended=use_extended)])

    def is_linear(self, num_cands):
        """Return True when the assignment of utilities is a linear order of ``num_cands`` candidates. 
        """

        return self.ranking().is_linear(num_cands=num_cands)

    def represents_ranking(self, r, use_extended=False): 
        """Return True when the utility represents the ranking ``r``."""
        
        if use_extended: 
            for x in r.cands: 
                for y in r.cands: 
                    if r.extended_strict_pref(x, y) and not self.extended_strict_pref(x, y): 
                        return False
                    elif r.extended_indiff(x, y) and not self.extended_indiff(x, y): 
                        return False

        else:
            for x in r.cands: 
                if not self.has_utility(x):
                    return False
            for x in r.cands: 
                for y in r.cands: 
                    if r.strict_pref(x, y) and not self.strict_pref(x, y): 
                        return False
                    elif r.indiff(x, y) and not self.indiff(x, y): 
                        return False
        return True
    
    def transformation(self, func): 
        """
        Return a new utility function that is the transformation of this utility function by the function ``func``.        
        """
        return Utility({
            x: func(self.val(x)) for x in self.defined_domain
        }, 
        domain = self.domain, 
        cmap=self.cmap)
    
    def linear_transformation(self, a=1, b=0): 
        """Return a linear transformation of the utility function: ``a * u(x) + b``.
        """
        
        lin_func = lambda x: a * self.val(x) + b
        return self.transformation(lin_func)
    
    def normalize_by_range(self): 
        """Return a normalized utility function.  Applies the *Kaplan* normalization to the utility function: 
        The new utility of an element x of the domain is (u(x) - min({u(x) | x in the domain})) / (max({u(x) | x in the domain })).
        """

        max_util = max(self.range)
        min_util = min(self.range)
        
        if max_util == min_util: 
            return Utility(
                {x:0 for x in self.defined_domain}, 
                domain = self.domain, 
                cmap=self.cmap)
        else: 
            return Utility({
                x: (self.val(x) - min_util) / (max_util - min_util) for x in self.defined_domain
            }, 
            domain = self.domain, 
            cmap=self.cmap)
        
    def normalize_by_standard_score(self): 
        """Replace each utility value with its standard score.  The standard score of a value is the number of standard deviations it is above the mean.
        """

        utility_values = np.array(list(self.image()))
    
        mean_utility = np.mean(utility_values)
        std_dev_utility = np.std(utility_values)

        return Utility({
                x: (self.val(x) - mean_utility) / std_dev_utility for x in self.defined_domain
            }, 
            domain = self.domain, 
            cmap=self.cmap)
        
    def expectation(self, prob):
        """Return the expected utility given a probability distribution ``prob``."""

        assert all([x in self.domain for x in prob.keys()]), "The domain of the probability distribution must be a subset of the domain of the utility function."

        return sum([prob[x] * self.util(x) for x in self.domain if x in prob.keys() and self.has_utility(x)])

    @classmethod
    def from_linear_ranking(cls, ranking, seed=None):
        """
        Return a utility function that represents the linear ranking.

        Parameters:
        ranking (List[int]): A list representing the linear ranking.
        seed (Optional[int]): An optional seed for random number generation.

        Returns:
        Utility: An instance of the Utility class.
        """

        if not (isinstance(ranking, list) or isinstance(ranking, tuple)):
            raise ValueError("Ranking must be a list.")
        if not len(set(ranking)) == len(ranking):
            raise ValueError("Ranking must be a list of unique numbers.")

        num_cands = len(ranking)
        rng = np.random.default_rng(seed)
        
        utilities = sorted(rng.random(size=num_cands), reverse=True)
        u_dict = {c: u for c, u in zip(ranking, utilities)}

        return cls(u_dict)

    def __str__(self):
        return self.display_str("U")

######
# Grade Functions
######

class Grade(_Mapping):
    def __init__(
        self, 
        grade_map, 
        grades, 
        candidates=None, 
        cmap=None, 
        gmap=None, 
        compare_function=None):
        """Constructor method for a Grade function."""

        assert all([g in grades for g in grade_map.values()]), f"All the grades in the grade map {grade_map} must be in the grades {grades}"
        assert candidates is None or all([x in candidates for x in grade_map.keys()]), f"The candidates {candidates} must contain all elements in the grade map {grade_map}"

        self.candidates = sorted(candidates) if candidates is not None else sorted(list(grade_map.keys())) 

        self.cmap = {x:str(x) for x in self.candidates} if cmap is None else cmap

        self.grades = grades 

        self.gmap = {g:str(g) for g in self.grades} if gmap is None else gmap

        super().__init__(
            grade_map, 
            domain=self.candidates, 
            codomain=self.grades, 
            item_map=self.cmap, 
            val_map=self.gmap,
            compare_function=compare_function)

    @property
    def graded_candidates(self):
        """Returns a list of the items that are assigned a grade."""
        return self.defined_domain

    def candidates_with_grade(self, g):
        """Returns a list of the items that are assigned the grade ``g``."""
        return self.inverse_image(g)

    def has_grade(self, x): 
        """Return True if x has a grade."""
        return self.has_value(x)
    
    def remove_cand(self, x):
        """Returns a grade function with the item ``x`` removed."""

        new_grades  = {y: self.val(y) for y in self.defined_domain if y != x}
        new_candidates = [y for y in self.domain if y != x]
        new_cmap   = {y: self.cmap[y] for y in self.cmap.keys() if y != x}
        return Grade(new_grades, grades=self.grades, candidates=new_candidates, cmap=new_cmap, gmap=self.gmap, compare_function=self.compare_function)
    
    def ranking(self): 
        """Return the ranking generated by this grade function."""

        return Ranking(
            {x:idx + 1 for idx, indiff_class in enumerate(self.sorted_domain()) 
             for x in indiff_class})
    
    def extended_ranking(self): 
        """Return the ranking generated by this grade function."""

        return Ranking(
            {x:idx + 1 for idx, indiff_class in enumerate(self.sorted_domain(extended=True))
             for x in indiff_class})

    def has_tie(self, use_extended=False): 
        """Return True when the utility has a tie."""
        return any([len(cs) != 1 for cs in self.sorted_domain(extended=use_extended)])

    def is_linear(self, num_cands):
        """Return True when the assignment of grades is a linear order of ``num_cands`` candidates. 
        """

        return self.ranking().is_linear(num_cands=num_cands)
    
    def __str__(self):
        return self.display_str("grade")





'''
    File: voting_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: November 6, 2021
    Update: January 15, 2023
    
    The VotingMethod class and helper functions for voting methods
'''

import functools
import inspect
import numpy as np
from numba import jit # Remove until numba supports python 3.11
import random
import json
from pref_voting.voting_method_properties import VotingMethodProperties
from filelock import FileLock, Timeout
import importlib.resources

import glob
import os

class VotingMethod(object): 
    """
    A class to add functionality to voting methods. 

    Args:
        vm (function): An implementation of a voting method. The function should accept a Profile, ProfileWithTies, MajorityGraph, and/or MarginGraph, and a keyword parameter ``curr_cands`` to find the winner after restricting to ``curr_cands``. 
        name (string): The Human-readable name of the voting method.
        properties (VotingMethodProperties): The properties of the voting method.
        input_types (list): The types of input that the voting method can accept.

    """
    def __init__(self, 
                 vm, 
                 name=None, 
                 input_types=None, 
                 skip_registration=False,                properties_file=None): 
        
        self.vm = vm
        self.name = name

        # Determine the path to the properties file
        if properties_file is None:

            properties_file = importlib.resources.files('pref_voting') / 'data' / 'voting_methods_properties.json'


        # Get the properties of the voting method
        try:
            with open(properties_file, "r") as file:
                vm_props = json.load(file)
        except FileNotFoundError:
            vm_props = {}
        except Exception as e:
            print(f"An error occurred while opening the properties file: {e}")
            vm_props = {}

        if name in vm_props:
            properties = VotingMethodProperties(**vm_props[name])
        else:
            properties = VotingMethodProperties()

        self.properties = properties
        self.input_types = input_types
        self.skip_registration = skip_registration
        self.algorithm = None

        functools.update_wrapper(self, vm)   

    def __call__(self, edata, curr_cands = None, **kwargs):
        
        if (curr_cands is not None and len(curr_cands) == 0) or len(edata.candidates) == 0: 
            return []
        
        # Set the algorithm from self.algorithm if it's not already provided in kwargs
        if 'algorithm' not in kwargs and self.algorithm is not None:
            params = inspect.signature(self.vm).parameters
            if 'algorithm' in params and params['algorithm'].kind in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD]:
                kwargs['algorithm'] = self.algorithm

        return self.vm(edata, curr_cands=curr_cands, **kwargs)

    def set_algorithm(self, algorithm):
        """
        Set the algorithm for the voting method if 'algorithm' is an accepted keyword parameter.

        Args:
            algorithm: The algorithm to set for the voting method.
        """
        params = inspect.signature(self.vm).parameters
        if 'algorithm' in params and params['algorithm'].kind in [inspect.Parameter.KEYWORD_ONLY, inspect.Parameter.POSITIONAL_OR_KEYWORD]:
            self.algorithm = algorithm
        else:
            raise ValueError(f"The method {self.name} does not accept 'algorithm' as a parameter.")
    
    def choose(self, edata, curr_cands = None): 
        """
        Return a randomly chosen element from the winning set. 
        """

        ws = self.__call__(edata, curr_cands = curr_cands)
        return random.choice(ws)
    
    def prob(self, edata, curr_cands = None): 
        """
        Return a dictionary representing the even-chance tiebreaking for the voting method.
        """
        
        ws = self.__call__(edata, curr_cands = curr_cands)
        return {c: 1.0 / len(ws) if c in ws else 0.0 for c in edata.candidates}
    
    def display(self, edata, curr_cands = None, cmap = None, **kwargs): 
        """
        Display the winning set of candidates.
        """

        cmap = cmap if cmap is not None else edata.cmap

        ws = self.__call__(edata, curr_cands = curr_cands, **kwargs)

        if ws is None:  # some voting methods, such as ``ranked_pairs_with_test``, may return None if it is taking long to compute the winner.
            print(f"{self.name} winning set is not available")
        else: 
            w_str = f"{self.name} winner is " if len(ws) == 1 else f"{self.name} winners are "
            print(w_str + "{" + ", ".join([str(cmap[c]) for c in ws]) + "}")
        
    def set_name(self, new_name):
        """Set the name of the voting method."""

        self.name = new_name

    def add_property(self, prop, value):
        """Add a property to the voting method."""

        setattr(self.properties, prop, value)

    def remove_property(self, prop):
        """Remove a property from the voting method."""

        delattr(self.properties, prop)

    def load_properties(self, filename=None):
        """Load the properties of the voting method from a JSON file."""
        
        # Determine the path to the properties file
        if filename is None:
            filename = importlib.resources.files('pref_voting') / 'data' / 'voting_methods_properties.json'
        lock = FileLock(f"{filename}.lock")
        with lock:
            try:
                with open(filename, 'r') as file:
                    vm_props = json.load(file)
            except FileNotFoundError:
                vm_props = {}

            if self.name in vm_props:
                self.properties = VotingMethodProperties(**vm_props[self.name])
            else:
                self.properties = VotingMethodProperties()

    def has_property(self, prop):
        """Check if the voting method has a property."""

        return self.properties[prop]
    
    def get_properties(self): 
        """Return the properties of the voting method."""
        
        return {
            "satisfied": [prop 
                          for prop, val in self.properties.items() 
                          if val is True],
            "violated": [prop 
                         for prop, val in self.properties.items() 
                         if val is False],
            "na": [prop 
                   for prop, val in self.properties.items() 
                   if val is None]
                }

    def save_properties(self, filename=None, timeout=10):
        """Save the properties of the voting method to a JSON file."""

        # Determine the path to the properties file
        if filename is None:
            filename = importlib.resources.files('pref_voting') / 'data' / 'voting_methods_properties.json'


        lock = FileLock(f"{filename}.lock", timeout=timeout)
        try:
            with lock:
                try:
                    with open(filename, 'r') as file:
                        vm_props = json.load(file)
                except FileNotFoundError:
                    vm_props = {}

                vm_props[self.name] = self.properties.__dict__

                with open(filename, 'w') as file:
                    json.dump(vm_props, file, indent=4, sort_keys=True)
        except Timeout:
            print(f"Could not acquire the lock within {timeout} seconds.")

    def get_violation_witness(self, prop, minimal_resolute=False, minimal=False):
        """Return the election that witnesses a violation of prop."""

        from pref_voting.profiles import Profile

        elections = {
            "minimal resolute": None,
            "minimal": None,
            "any": None
        }
        if self.properties[prop]:
            print(f"{self.name} satisfies {prop}, no election returned.")
            return elections
        elif self.properties[prop] is None:
            print(f"{self.name} does not have a value for {prop}, no election returned.")
            return elections
        else:
            dir = importlib.resources.files('pref_voting') / 'data' / 'examples' / prop

            for f in glob.glob(f"{dir}*"):
                fname = os.path.basename(f)
                is_min = fname.startswith("minimal_")
                is_min_resolute = fname.startswith("minimal_resolute")
                found_it = False
                if is_min_resolute and fname.startswith(f"minimal_resolute_{self.name.replace(' ', '_')}"): 
                    print(f"Minimal resolute election for a violation of {prop} found.")
                    elections["minimal resolute"] = Profile.from_preflib(f)
                if is_min and not is_min_resolute and fname.startswith(f"minimal_{self.name.replace(' ', '_')}"):
                    print(f"Minimal election for a violation of {prop} found.")
                    elections["minimal"] = Profile.from_preflib(f)

                elif not is_min and not is_min_resolute and  fname.startswith(f"{self.name.replace(' ', '_')}"):
                    elections["any"] = Profile.from_preflib(f)
            if all([v is None for v in elections.values()]):
                print(f"No election found illustrating the violation of {prop}.")
            return elections

    def check_property(self, prop, include_counterexample=True): 
        """Check if the voting method satisfies a property."""
        from pref_voting.axioms import axioms_dict

        if not self.properties[prop]: 
            print(f"{self.name} does not satisfy {prop}")
            if include_counterexample:
                if prop in axioms_dict:
                    #prof = prof
                    axioms_dict[prop].counterexample(self)

        elif self.properties[prop] is None: 
            print(f"{self.name} does not have a value for {prop}")
        
        else:
            print(f"{self.name} satisfies {prop}")

    def __str__(self): 
        return f"{self.name}"

def vm(name=None, 
       input_types=None, 
       skip_registration=False):
    """
    A decorator used when creating a voting method. 
    """
    def wrapper(f):
        return VotingMethod(f, 
                            name=name,
                            input_types=input_types, 
                            skip_registration=skip_registration)
    return wrapper

@jit(nopython=True, fastmath=True)
def isin(arr, val):
    """compiled function testing if the value val is in the array arr
    """
    
    for i in range(arr.shape[0]):
        if (arr[i]==val):
            return True
    return False

@jit(nopython=True, fastmath=True)
def _num_rank_first(rankings, rcounts, cands_to_ignore, cand):
    """The number of voters that rank candidate cand first after ignoring the candidates in 
    cands_to_ignore
    
    Parameters
    ----------
    rankings:  2d numpy array
        list of linear orderings of the candidates  
    rcounts:  1d numpy array
        list of numbers of voters with the rankings  
    cands_to_ignore:   1d numpy array
        list of the candidates to ignore
    cand: int
        a candidate
    
    Key assumptions: 
        * the candidates are named 0...num_cands - 1, and c1 and c2 are 
          numbers between 0 and num_cands - 1
        * voters submit linear orders over the candidate        
    """
    
    num_voters = len(rankings)
    
    top_cands_indices = np.zeros(num_voters, dtype=np.int32)
    
    for vidx in range(num_voters): 
        for level in range(0, len(rankings[vidx])):
            if not isin(cands_to_ignore, rankings[vidx][level]):
                top_cands_indices[vidx] = level
                break                
    top_cands = np.array([rankings[vidx][top_cands_indices[vidx]] for vidx in range(num_voters)])
    is_cand = top_cands == cand # set to 0 each candidate not equal to cand
    return np.sum(is_cand * rcounts) 


@jit(nopython=True, fastmath=True)
def _num_rank_last(rankings, rcounts, cands_to_ignore, cand):
    """The number of voters that rank candidate cand last after ignoring the candidates in 
    cands_to_ignore
    
    Parameters
    ----------
    rankings:  2d numpy array
        list of linear orderings of the candidates  
    rcounts:  1d numpy array
        list of numbers of voters with the rankings  
    cands_to_ignore:   1d numpy array
        list of the candidates to ignore
    cand: int
        a candidate
    
    Key assumptions: 
        * the candidates are named 0...num_cands - 1, and c1 and c2 are 
          numbers between 0 and num_cands - 1
        * voters submit linear orders over the candidate        
    """
    
    num_voters = len(rankings)
    
    last_cands_indices = np.zeros(num_voters, dtype=np.int32)
    
    for vidx in range(num_voters): 
        for level in range(len(rankings[vidx]) - 1,-1,-1):
            if not isin(cands_to_ignore, rankings[vidx][level]):
                last_cands_indices[vidx] = level
                break                
    bottom_cands = np.array([rankings[vidx][last_cands_indices[vidx]] for vidx in range(num_voters)])
    is_cand = bottom_cands  == cand
    return np.sum(is_cand * rcounts) 



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pref_voting', 'pref_voting.io']

package_data = \
{'': ['*'],
 'pref_voting': ['data/examples/condorcet_winner/*',
                 'data/voting_methods_properties.json',
                 'data/voting_methods_properties.json',
                 'data/voting_methods_properties.json.lock',
                 'data/voting_methods_properties.json.lock']}

install_requires = \
['filelock>=3.12.2,<4.0.0',
 'matplotlib>=3.5.2,<4.0.0',
 'nashpy>=0.0.40,<0.0.41',
 'networkx>=3.0,<4.0',
 'numba>=0.61.0,<0.62.0',
 'ortools>=9.8.0,<10.0.0',
 'pathos>=0.3.3,<0.4.0',
 'preflibtools>=2.0.22,<3.0.0',
 'prefsampling>=0.1.16,<0.2.0',
 'random2>=1.0.1,<2.0.0',
 'scipy>=1.0.0,<2.0.0',
 'seaborn>=0.13.2,<0.14.0',
 'tabulate>=0.9.0,<0.10.0']

setup_kwargs = {
    'name': 'pref-voting',
    'version': '1.16.27',
    'description': 'pref_voting is a Python package that contains tools to reason about elections and margin graphs, and implementations of voting methods.',
    'long_description': 'pref_voting\n==========\n[![DOI](https://joss.theoj.org/papers/10.21105/joss.07020/status.svg)](https://doi.org/10.21105/joss.07020) [![DOI](https://zenodo.org/badge/578984957.svg)](https://doi.org/10.5281/zenodo.14675583)\n\n[![Tests](https://github.com/voting-tools/pref_voting/actions/workflows/tests.yml/badge.svg)](https://github.com/voting-tools/pref_voting/actions/workflows/tests.yml)\n\n\n> [!NOTE]\n> - [**Documentation**](https://pref-voting.readthedocs.io/)\n> - [**Installation**](https://pref-voting.readthedocs.io/en/latest/installation.html)  \n> - [**Example Notebooks**](https://github.com/voting-tools/pref_voting/tree/main/examples)  \n> - [**Example Elections**](https://github.com/voting-tools/election-analysis)\n> - [**► pref_voting web app**](https://pref.tools/pref_voting/)\n\nSee the [COMSOC community page](https://comsoc-community.org/tools) for an overview of other software tools related to Computational Social Choice.\n\n## Installation\n\nThe package can be installed using the ``pip3`` package manager:\n\n```bash\npip3 install pref_voting\n```\n**Notes**: \n* The package requires Python 3.10 or higher and has been tested on Python 3.12.\n\n* Since the package uses Numba, refer to the [Numba documentation for the latest supported Python version](https://numba.readthedocs.io/en/stable/user/installing.html#version-support-information).\n* If you have both Python 2 and Python 3 installed on your system, make sure to use ``pip3`` instead of pip to install packages for Python 3. Alternatively, you can use ``python3 -m pip`` to ensure you\'re using the correct version of pip. If you have modified your system\'s defaults or soft links, adjust accordingly.\n\nSee the [installation guide](https://pref-voting.readthedocs.io/en/latest/installation.html) for more detailed instructions.\n\n## Example Usage\n\nA profile (of linear orders over the candidates) is created by initializing a `Profile` class object.  Simply provide a list of rankings (each ranking is a tuple of numbers) and a list giving the number of voters with each ranking:\n\n```python\nfrom pref_voting.profiles import Profile\n\nrankings = [\n    (0, 1, 2, 3), # candidate 0 is ranked first, candidate 1 is ranked second, candidate 2 is ranked 3rd, and candidate 3 is ranked last.\n    (2, 3, 1, 0), \n    (3, 1, 2, 0), \n    (1, 2, 0, 3), \n    (1, 3, 2, 0)]\n\nrcounts = [5, 3, 2, 4, 3] # 5 voters submitted the first ranking (0, 1, 2, 3), 3 voters submitted the second ranking, and so on.\n\nprof = Profile(rankings, rcounts=rcounts)\n\nprof.display() # display the profile\n```\n\nThe function `generate_profile` is used to generate a profile for a given number of candidates and voters:  \n\n```python\nfrom pref_voting.generate_profiles import generate_profile\n\n# generate a profile using the Impartial Culture probability model\nprof = generate_profile(3, 4) # prof is a Profile object with 3 candidates and 4 voters\n\n# generate a profile using the Impartial Anonymous Culture probability model\nprof = generate_profile(3, 4, probmod = "IAC") # prof is a Profile object with 3 candidates and 4 voters \n```\n\nThe `Profile` class has a number of methods that can be used to analyze the profile. For example, to determine the margin of victory between two candidates, the plurality scores, the Copeland scores, the Borda scores, the Condorcet winner, the weak Condorcet winner, and the Condorcet loser, and whether the profile is uniquely weighted, use the following code:\n\n```python\n\nprof = Profile([\n    [2, 1, 0, 3], \n    [3, 2, 0, 1], \n    [3, 1, 0, 2]], \n    rcounts=[2, 2, 3])\n\nprof.display()\n\nprint(f"The margin of 1 over 3 is {prof.margin(1, 3)}")\nprint(f"The Plurality scores are {prof.plurality_scores()}")\nprint(f"The Copeland scores are {prof.copeland_scores()}")\nprint(f"The Borda scores are {prof.borda_scores()}")\nprint(f"The Condorcet winner is {prof.condorcet_winner()}")\nprint(f"The weak Condorcet winner is {prof.weak_condorcet_winner()}")\nprint(f"The Condorcet loser is {prof.condorcet_loser()}")\nprint(f"The profile is uniquely weighted: {prof.is_uniquely_weighted()}")\n\n```\n\nTo use one of the many voting methods, import the function from `pref_voting.voting_methods` and apply it to the profile: \n\n```python\nfrom pref_voting.generate_profiles import generate_profile\nfrom pref_voting.voting_methods import *\n\nprof = generate_profile(3, 4) # create a profile with 3 candidates and 4 voters\nsplit_cycle(prof) # returns the sorted list of winning candidates\nsplit_cycle.display(prof) # displays the winning candidates\n\n```\n\nAdditional notebooks that demonstrate how to use the package can be found in the [examples directory](https://github.com/voting-tools/pref_voting/tree/main/examples)\n\nSome interesting political elections are analyzed using pref_voting in the [election-analysis repository](https://github.com/voting-tools/election-analysis).\n\nConsult the documentation [https://pref-voting.readthedocs.io](https://pref-voting.readthedocs.io) for a complete overview of the package. \n\n\n## Testing\n \nTo ensure that the package is working correctly, you can run the test suite using [pytest](https://docs.pytest.org/en/stable/). The test files are located in the `tests` directory. Follow the instructions below based on your setup.\n\n### Prerequisites\n\n- **Python 3.9 or higher**: Ensure you have a compatible version of Python installed.\n- **`pytest`**: Install `pytest` if it\'s not already installed.\n\n### Running the tests\n\nIf you are using **Poetry** to manage your dependencies, run the tests with:\n\n```bash\npoetry run pytest\n\n```\n \nFrom the command line, run:\n\n```bash\npytest\n```\n\nFor more detailed output, add the -v or --verbose flag:\n\n```bash\npytest -v\n```\n\n## How to cite\n \nIf you would like to acknowledge our work in a scientific paper,\nplease use the following citation:\n\nWesley H. Holliday and Eric Pacuit (2025). pref_voting: The Preferential Voting Tools package for Python. Journal of Open Source Software, 10(105), 7020. https://doi.org/10.21105/joss.07020\n\n### BibTeX:\n\n```bibtex\n@article{HollidayPacuit2025, \n  author = {Wesley H. Holliday and Eric Pacuit}, \n  title = {pref_voting: The Preferential Voting Tools package for Python}, \n  journal = {Journal of Open Source Software},\n  year = {2025}, \n  publisher = {The Open Journal}, \n  volume = {10}, \n  number = {105}, \n  pages = {7020}, \n  doi = {10.21105/joss.07020}\n}\n\n```\n\nAlternatively, you can cite the archived code repository\nat [zenodo](https://doi.org/10.5281/zenodo.14675583).\n\n## Contributing\n\nIf you would like to contribute to the project, please see the [contributing guidelines](CONTRIBUTING.md).\n\n## Questions?\n\nFeel free to [send an email](https://pacuit.org/) if you have questions about the project.\n\n## License\n\n[MIT](https://github.com/voting-tools/pref_voting/blob/main/LICENSE.txt)\n',
    'author': 'Eric Pacuit',
    'author_email': 'epacuit@umd.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/voting-tools/pref_voting',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import numpy as np
from pref_voting.spatial_profiles import SpatialProfile


def generate_covariance(n_dimensions, std, rho):
    """
    Generates a covariance matrix for a multivariate normal distribution with the given standard deviation and correlation coefficient.

    Parameters
    ----------
    n_dimensions : int
        The number of dimensions.
    std : float
        The standard deviation.
    rho : float
        The correlation coefficient.

    Returns
    -------
    cov : numpy.ndarray
        The covariance matrix for a multivariate normal distribution.
    """
    assert std > 0, "Standard deviation must be positive"
    assert rho >= 0 and rho <= 1, "Correlation coefficient must be between 0 and 1"
    assert n_dimensions > 0, "Number of dimensions must be positive"

    cov = (std**2) * (np.eye(n_dimensions) + rho * (np.ones((n_dimensions, n_dimensions)) - np.eye(n_dimensions)))
    
    return cov

def generate_spatial_profile(num_cands, num_voters, num_dims, cand_cov = None, voter_cov = None, num_profiles = 1): 

    """ 
    Generates a spatial profile with the candidate and voter positions generated by a multivariate normal distribution with num_dims dimensions and cand_cov the covariance matrix for candidates and voter_cov the covariance matrix for voters.

    Parameters
    ----------
    num_cands : int
        The number of candidates.
    num_voters : int
        The number of voters.
    num_dims : int  
        The number of dimensions.
    cand_cov : numpy.ndarray, optional
        The covariance matrix for the multivariate normal distribution for candidates. The default is None.
    voter_cov : numpy.ndarray, optional
        The covariance matrix for the multivariate normal distribution for voters. The default is the identity matrix.
    
    Returns
    -------
    SpatialProfile  
        A spatial profile with the candidate and voter positions generated by a multivariate normal distribution with num_dims dimensions and cand_cov the covariance matrix for candidates and voter_cov the covariance matrix for voters.

    """
    cand_cov = np.eye(num_dims) if cand_cov is None else cand_cov
    voter_cov = np.eye(num_dims) if voter_cov is None else voter_cov

    assert num_dims == cand_cov.shape[0] == cand_cov.shape[1], "Candidate covariance matrix must be square and have the same number of dimensions as the number of dimensions specified"
    assert num_dims == voter_cov.shape[0] == voter_cov.shape[1], "Voter covariance matrix must be square and have the same number of dimensions as the number of dimensions specified"

    cand_mean = np.zeros(num_dims)
    voter_mean = np.zeros(num_dims)

    cand_samples = np.random.multivariate_normal(cand_mean, cand_cov, size=(num_profiles,num_cands))
    voter_samples = np.random.multivariate_normal(voter_mean, voter_cov, size=(num_profiles,num_voters))

    profs = [SpatialProfile({c: cand_samples[pidx][c] for c in range(num_cands)},
                            {v: voter_samples[pidx][v] for v in range(num_voters)})
                          for pidx in range(num_profiles)]

    return profs[0] if num_profiles == 1 else profs

def generate_spatial_profile_polarized(
    cand_clusters, 
    voter_clusters, 
    cluster_types = None, 
    num_profiles = 1):
    """
    Generates a spatial profile with polarized clusters of candidates and voters.   
    
    Parameters
    ----------
    cand_clusters : list
        A list of tuples of the form (mean, covariance, number of candidates) for each cluster of candidates.
    voter_clusters : list
        A list of tuples of the form (mean, covariance, number of voters) for each cluster of voters.
    cluster_types : dict, optional
        A list of the same length as cand_cluster that associates each cluster to the type of candidate. The default is None.

    num_profiles : int, optional
        The number of profiles to generate. The default is 1.

    Returns
    -------
    SpatialProfile
        A spatial profile with polarized clusters of candidates and voters.
    """

    cluster_types = cluster_types if cluster_types is not None else list(range(len(cand_clusters)))
    cand_samples = list()
    candidate_types = {}
    total_num_cands = 0
    for cluster_idx, cand_cluster in enumerate(cand_clusters):
        cand_mean, cand_cov, num_cands = cand_cluster
        candidate_types.update({cidx: cluster_types[cluster_idx] for cidx in range(total_num_cands, total_num_cands+num_cands)})
        total_num_cands += num_cands
        cluster_samples = np.random.multivariate_normal(cand_mean, cand_cov, 
                                                        size=(num_profiles,num_cands))
        if len(cand_samples) == 0:
            cand_samples = cluster_samples
        else:
            cand_samples = np.concatenate([cand_samples, cluster_samples], axis=1)

    voter_samples = list()
    total_num_voters = 0
    for voter_cluster in voter_clusters:
        voter_mean, voter_cov, num_voters = voter_cluster
        total_num_voters += num_voters
        cluster_samples = np.random.multivariate_normal(voter_mean, voter_cov, 
                                                        size=(num_profiles,num_voters))
        if len(voter_samples) == 0:
            voter_samples = cluster_samples
        else:
            voter_samples = np.concatenate([voter_samples, cluster_samples], axis=1) 

    
    profs = [SpatialProfile({cidx: cand_samples[pidx][cidx]
                           for cidx in range(total_num_cands)},
                           {vidx: voter_samples[pidx][vidx]
                            for vidx in range(total_num_voters)}, 
                            candidate_types=candidate_types) 
                            for pidx in range(num_profiles)]
    
    return profs[0] if num_profiles == 1 else profs


def generate_spatial_profile_polarized_cands_randomly_polarized_voters(
        cand_clusters, 
        num_voters,
        voter_distributions,
        cluster_types = None, 
        num_profiles = 1):
    """
    Generates a spatial profile with polarized clusters of candidates and voters.   
    
    Parameters
    ----------
    cand_clusters : list
        A list of tuples of the form (mean, covariance, number of candidates) for each cluster of candidates.
    num_voters : int
        The number of voters.
    voter_distributions : list
        A list of tuples of the form (mean, covariance, prob) for each distribution of voters, where prob is the probability that a voter is assigned to this cluster.
    cluster_types : dict, optional
        A list of the same length as cand_cluster that associates each cluster to the type of candidate. The default is None.
    num_profiles : int, optional
        The number of profiles to generate. The default is 1.

    Returns
    -------
    SpatialProfile
        A spatial profile with polarized clusters of candidates and voters.
    """

    cluster_types = cluster_types if cluster_types is not None else list(range(len(cand_clusters)))
    cand_samples = list()
    candidate_types = {}
    total_num_cands = 0
    for cluster_idx, cand_cluster in enumerate(cand_clusters):
        cand_mean, cand_cov, num_cands = cand_cluster
        candidate_types.update({cidx: cluster_types[cluster_idx] for cidx in range(total_num_cands, total_num_cands+num_cands)})
        total_num_cands += num_cands
        cluster_samples = np.random.multivariate_normal(cand_mean, cand_cov, size=(num_profiles,num_cands))
        if len(cand_samples) == 0:
            cand_samples = cluster_samples
        else:
            cand_samples = np.concatenate([cand_samples, cluster_samples], axis=1)

    all_potential_voters = list()
    voter_cluster_probs = list()
    for voter_distribution in voter_distributions:
        voter_mean, voter_cov, prob = voter_distribution
        all_potential_voters.append(np.random.multivariate_normal(voter_mean, voter_cov, size=(num_profiles, num_voters)))
        voter_cluster_probs.append(prob)

    # for each profile, randomly assign voters to clusters

    #voters = range(num_voters)

    voters_clusters_assignment = np.random.choice(len(all_potential_voters), size=(num_profiles, num_voters), p = voter_cluster_probs)
    profs = [SpatialProfile({cidx: cand_samples[pidx][cidx]
                           for cidx in range(total_num_cands)},
                           {vidx: all_potential_voters[voters_clusters_assignment[pidx][vidx]][pidx][vidx]
                            for vidx in range(num_voters)}, 
                            candidate_types=candidate_types) 
                            for pidx in range(num_profiles)]
    
    return profs[0] if num_profiles == 1 else profs


'''
    File: other_methods.py
    Author: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)
    Date: January 12, 2022
    Updated: April 21, 2025

'''
from pref_voting.voting_method import *
from pref_voting.scoring_methods import plurality
from pref_voting.profiles import _find_updated_profile, _num_rank
from pref_voting.profiles_with_ties import ProfileWithTies
from pref_voting.weighted_majority_graphs import MarginGraph
from itertools import combinations, permutations
from pref_voting.voting_method_properties import ElectionTypes
from pref_voting.rankings import Ranking
from pref_voting.social_welfare_function import swf
import numpy as np
from pref_voting.profiles_with_ties import _num_rank_profile_with_ties
import copy
from ortools.linear_solver import pywraplp

@vm(name = "Absolute Majority",
    skip_registration=True, # skip registration since aboslute majority may return an empty list
    input_types = [ElectionTypes.PROFILE])
def absolute_majority(profile, curr_cands = None):
    """The absolute majority winner is the candidate with a strict majority  of first place votes.  Returns an empty list if there is no candidate with a strict majority of first place votes. Otherwise returns the absolute majority winner in the ``profile`` restricted to ``curr_cands``.

    ..note:
        The term 'absolute majority' for this voting method comes from Charles Dodgson's famous pamplet of 1873, "A Discussion of the Various Methods of Procedure in Conducting Elections" (see I. McLean and A. Urken, *Classics of Social Choice*, 1995, p. 281, or A. D. Taylor, "Social Choice and the Mathematics of Manipulation," 2005, p. 11).

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    .. important:: 
        Formally, this is *not* a voting method since the function might return an empty list (when there is no candidate with a strict majority of first place votes).  Also, if there is an absolute majority winner, then that winner is unique. 

    :Example:
        .. exec_code::

            from pref_voting.profiles import Profile
            from pref_voting.other_methods import absolute_majority
            
            prof1 = Profile([[0, 1, 2], [1, 0, 2], [2, 1, 0]], [3, 1, 2])
            prof1.display()
            absolute_majority.display(prof1)

            prof2 = Profile([[0, 1, 2], [1, 0, 2], [1, 2, 0]], [5, 1, 2])
            prof2.display()
            absolute_majority.display(prof2)

    """
    maj_size = profile.strict_maj_size()
    curr_cands = profile.candidates if curr_cands is None else curr_cands

    plurality_scores = profile.plurality_scores(curr_cands = curr_cands)
    abs_maj_winner = [c for c in curr_cands if plurality_scores[c] >= maj_size]

    return sorted(abs_maj_winner)

@vm(name = "Pareto",
    input_types = [ElectionTypes.PROFILE, ElectionTypes.PROFILE_WITH_TIES])
def pareto(profile, curr_cands = None, strong_Pareto = False, use_extended_strict_preferences = True):
    """Returns the set of candidates who are not Pareto dominated.

    For ProfilesWithTies, if strong_Pareto == True, then a dominates b if some voter strictly prefers a to b and no voter strictly prefers b to a.

    Args:
        prof (Profile, ProfileWithTies): An anonymous profile of linear (or strict weak) orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    """

    if type(profile) == ProfileWithTies:
        currently_using_extended_strict_preferences = profile.using_extended_strict_preference
        if use_extended_strict_preferences:
            profile.use_extended_strict_preference()
        
    Pareto_dominated = set()
    candidates = profile.candidates if curr_cands is None else curr_cands
    for a in candidates:
        for b in candidates:
            if not strong_Pareto and profile.support(a,b) == profile.num_voters:
                Pareto_dominated.add(b)

            if strong_Pareto and profile.support(a,b) > 0 and profile.support(b,a) == 0:
                Pareto_dominated.add(b)     

    if type(profile) == ProfileWithTies and use_extended_strict_preferences:
        if not currently_using_extended_strict_preferences:
            profile.use_strict_preference()

    return sorted(list(set(candidates) - Pareto_dominated))
    

## Kemeny-Young Method 
#
def kendalltau_dist(rank_a, rank_b):
    index_b = {c: i for i, c in enumerate(rank_b)}
    tau = 0
    for i, j in combinations(rank_a, 2):
        # by definition of itertools.combinations, index_a[i] < index_a[j]
        if index_b[i] > index_b[j]:
            tau += 1
    return tau

def kendalltau_dist_for_rankings_with_ties(
        candidates, 
        ranking1, 
        ranking2,
        penalty=0.5):

    tau = 0
    for c1, c2 in combinations(candidates, 2):
        # by definition of itertools.combinations, index_a[i] < index_a[j]
        
        if (ranking1.extended_strict_pref(c1, c2) and ranking2.extended_strict_pref(c2, c1)) or (ranking1.extended_strict_pref(c2, c1) and ranking2.extended_strict_pref(c1, c2)):
            tau += 1
        elif (ranking1.extended_strict_pref(c1, c2) and ranking2.extended_indiff(c1, c2)) or (ranking1.extended_strict_pref(c2, c1) and ranking2.extended_indiff(c1, c2)):
            tau += penalty
        elif (ranking1.extended_indiff(c1, c2) and ranking2.extended_strict_pref(c1, c2)) or (ranking1.extended_indiff(c1, c2) and ranking2.extended_strict_pref(c2, c1)) :
            tau += penalty

    return tau


def _kemeny_young_rankings(rankings, rcounts, candidates): 
    
    rankings_dist = dict()
    for ranking in permutations(candidates): 
        rankings_dist[tuple(ranking)] = sum(c * kendalltau_dist(tuple(r), ranking) 
        for r,c in zip(rankings, rcounts))
    min_dist = min(rankings_dist.values())

    lin_orders = [r for r in rankings_dist.keys() if rankings_dist[r] == min_dist]
    
    return lin_orders, min_dist

def kemeny_young_rankings(profile, curr_cands = None): 
    """
    A Kemeny-Young ranking is a ranking that minimizes the sum of the Kendall tau distances to the voters' rankings.  
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        rankings: A list of Kemeny-Young rankings.
        
        dist: The minimum distance of the Kemeny-Young rankings.


    :Example:
        .. exec_code::

            from pref_voting.profiles import Profile
            from pref_voting.other_methods import kemeny_young, kemeny_young_rankings
            
            prof1 = Profile([[0, 1, 2], [1, 0, 2], [2, 1, 0]], [3, 1, 2])
            prof1.display()
            kyrs, d = kemeny_young_rankings(prof1)
            print(f"Minimal distance: {d}")
            for kyr in kyrs: 
                print(f"ranking: {kyr}")

            prof2 = Profile([[0, 1, 2], [1, 0, 2], [1, 2, 0]], [5, 1, 2])
            prof2.display()
            kyrs, d = kemeny_young_rankings(prof2)
            print(f"Minimal distance: {d}")
            for kyr in kyrs: 
                print(f"ranking: {kyr}")

    """
    candidates = profile.candidates if curr_cands is None else curr_cands
    
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]),  profile.num_cands)
    return _kemeny_young_rankings(list(rankings), list(profile._rcounts), candidates)


@vm(name = "Kemeny-Young",
    input_types = [ElectionTypes.PROFILE,ElectionTypes.PROFILE_WITH_TIES,  ElectionTypes.MARGIN_GRAPH])
def kemeny_young(edata, curr_cands = None, algorithm = "marginal"): 
    """A Kemeny-Young ranking is a ranking that maximizes the sum of the margins of pairs of candidates in the ranking. Equivalently, a Kemeny-Young ranking is a ranking that minimizes the sum of the Kendall tau distances to the voters' rankings. The Kemeny-Young winners are the candidates that are ranked first by some Kemeny-Young ranking.

    Args:
        edata (Profile, ProfileWithTies, MarginGraph): Any election data that has a `margin` method.
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        algorithm (str, optional): The algorithm to use.  Options are "marginal" and "Kendall tau". If "marginal" is used, then the Kemeny-Young rankings are computed by finding the sum of the margins of each pair of candidates in the ranking.  If "Kendall tau" is used, then the Kemeny-Young rankings are computed by summing the Kendall tau distances to the voters' rankings.  Default is "marginal".

    Returns: 
        A sorted list of candidates

    :Example:

        .. exec_code::

            from pref_voting.profiles import Profile
            from pref_voting.other_methods import kemeny_young, kemeny_young_rankings
            
            prof1 = Profile([[0, 1, 2], [1, 0, 2], [2, 1, 0]], [3, 1, 2])
            prof1.display()
            kyrs, d = kemeny_young_rankings(prof1)
            print(f"Minimal distance: {d}")
            for kyr in kyrs: 
                print(f"ranking: {kyr}")
            kemeny_young.display(prof1)

            prof2 = Profile([[0, 1, 2], [1, 0, 2], [1, 2, 0]], [5, 1, 2])
            prof2.display()
            kyrs, d = kemeny_young_rankings(prof2)
            print(f"Minimal distance: {d}")
            for kyr in kyrs: 
                print(f"ranking: {kyr}")
            kemeny_young.display(prof2)

    """
    assert algorithm in ["marginal", "Kendall tau"], "Algorithm must be either 'marginal' or 'Kendall tau'."

    candidates = edata.candidates if curr_cands is None else curr_cands

    if isinstance(edata, MarginGraph) or isinstance(edata,ProfileWithTies):
        algorithm = "marginal"

    if algorithm == "Kendall tau":
        rankings = edata._rankings if curr_cands is None else _find_updated_profile(edata._rankings, np.array([c for c in edata.candidates if c not in curr_cands]), edata.num_cands)
        ky_rankings, min_dist = _kemeny_young_rankings(list(rankings), list(edata._rcounts), candidates)

    if algorithm == "marginal":

        best_ranking_score = 0
        ky_rankings = []

        for r in permutations(candidates):

            score_of_r = 0 
            for i in r[:-1]:
                for j in r[r.index(i)+1:]:
                    score_of_r += edata.margin(i,j)

            if score_of_r > best_ranking_score:
                best_ranking_score = score_of_r
                ky_rankings = [r]
            if score_of_r == best_ranking_score:
                ky_rankings.append(r)
    
    return sorted(list(set([r[0] for r in ky_rankings])))

@vm("Preliminary Weighted Condorcet",
    input_types = [ElectionTypes.PROFILE])
def preliminary_weighted_condorcet(prof, curr_cands = None, show_orders = False, require_positive_plurality_score = False):
    """The preliminary version of the Weighted Condorcet Rule in Tideman's book, Collective Decisions and Voting (p. 223). The winners are the candidates ranked first by some linear order of the candidates with highest score, where the score of an order (c_1,...,c_n) is the sum over all i<j of the margin of c_i vs. c_j multiplied by the plurality scores of c_i and c_j. 
    
    The multiplication by plurality scores is what distinguishes this method from the Kemeny-Young method.
    
    Tideman (p. 224) defines a more complicated Weighted Condorcet rule that is intended to be used when some candidates receive zero first-place votes.
    
    Args:
        prof (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        show_orders (bool): If True, then print the set of best orders.
        require_positive_plurality_score (bool): If True, then require that all candidates have a positive plurality score.

    Returns:
        A sorted list of candidates
    """

    cands = curr_cands if curr_cands is not None else prof.candidates

    if require_positive_plurality_score:
        assert all([prof.plurality_scores(curr_cands=curr_cands)[c] > 0 for c in cands]), "All candidates must have a positive plurality score."

    best_order_score = 0
    best_orders = []

    for r in permutations(cands):

        score_of_r = 0 
        for i in r[:-1]:
            for j in r[r.index(i)+1:]:
                score_of_r += (prof.plurality_scores(curr_cands=curr_cands)[i] * prof.plurality_scores(curr_cands=curr_cands)[j] * prof.margin(i,j))

        if score_of_r > best_order_score:
            best_order_score = score_of_r
            best_orders = [r]
        if score_of_r == best_order_score:
            best_orders.append(r)

    if show_orders == True:
        print(f"Best orders: {set(best_orders)}")

    winners = [r[0] for r in best_orders]

    return list(set(winners))

### Bucklin

@vm(name = "Bucklin",
    input_types = [ElectionTypes.PROFILE])
def bucklin(profile, curr_cands = None): 
    """If a candidate has a strict majority of first-place votes, then that candidate is the winner. If no such candidate exists, then check the candidates that are ranked first or second.  If a candidate has a strict majority of first- or second-place voters, then that candidate is the winner. If no such winner is found move on to the 3rd, 4th, etc. place votes.  Return the candidates with the greatest overall score.  
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.other_methods import bucklin

        prof = Profile([[1, 0, 2], [0, 2, 1], [0, 1, 2]], [2, 1, 1])

        prof.display()
        bucklin.display(prof)

    """
    strict_maj_size = profile.strict_maj_size()
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    
    num_cands = candidates
    
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), profile.num_cands)

    rcounts = profile._rcounts
    
    num_cands = len(candidates)
    ranks = range(1, num_cands + 1)
    
    cand_to_num_voters_rank = dict()
    for r in ranks:
        cand_to_num_voters_rank[r] = {c: _num_rank(rankings, rcounts, c, r)
                                      for c in candidates}
        cand_scores = {c:sum([cand_to_num_voters_rank[_r][c] for _r in cand_to_num_voters_rank.keys()]) 
                       for c in candidates}
        if any([s >= strict_maj_size for s in cand_scores.values()]):
            break
    max_score = max(cand_scores.values())
    return sorted([c for c in candidates if cand_scores[c] >= max_score])


def bucklin_with_explanation(profile, curr_cands = None): 
    """Return the Bucklin winners and the score for each candidate. 
 
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 

        A sorted list of candidates

        A dictionary assigning the score for each candidate. 

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.other_methods import bucklin_with_explanation

        prof = Profile([[1, 0, 2], [0, 2, 1], [0, 1, 2]], [2, 1, 1])

        prof.display()
        sb_ws, scores = bucklin_with_explanation(prof)

        print(f"The winners are {sb_ws}")
        print(f"The candidate scores are {scores}")

    """
    strict_maj_size = profile.strict_maj_size()
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    
    num_cands = candidates
    
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), profile.num_cands)

    rcounts = profile._rcounts
    
    num_cands = len(candidates)
    ranks = range(1, num_cands + 1)
    
    cand_to_num_voters_rank = dict()
    for r in ranks:
        cand_to_num_voters_rank[r] = {c: _num_rank(rankings, rcounts, c, r)
                                      for c in candidates}
        cand_scores = {c:sum([cand_to_num_voters_rank[_r][c] for _r in cand_to_num_voters_rank.keys()]) 
                       for c in candidates}
        if any([s >= strict_maj_size for s in cand_scores.values()]):
            break
    max_score = max(cand_scores.values())
    return sorted([c for c in candidates if cand_scores[c] >= max_score]), cand_scores

@vm(name="Bucklin for Truncated Linear Orders", 
    input_types=[ElectionTypes.PROFILE_WITH_TIES])
def bucklin_for_truncated_linear_orders(profile, curr_cands=None):
    """The Bucklin voting method adapted for truncated linear orders using ProfileWithTies.
    
    Args:
        profile: A ProfileWithTies object that represents an election profile.
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in curr_cands

    Returns: 
        A sorted list of candidates that win according to the Bucklin method.
        
    .. note::
        This is an adaptation of the Bucklin method for truncated linear orders. Empty ballots are removed before calculating
        the majority threshold, similar to how instant_runoff_for_truncated_linear_orders handles truncated profiles.
    """
    assert all([not r.has_overvote() for r in profile.rankings]), "Bucklin is only defined when all the ballots are truncated linear orders."

    # Remove empty rankings and get working copy of profile
    working_profile = copy.deepcopy(profile)
    working_profile.remove_empty_rankings()
    
    strict_maj_size = working_profile.strict_maj_size()
    candidates = working_profile.candidates if curr_cands is None else curr_cands
    
    if curr_cands is not None:
        working_profile = working_profile.remove_candidates([c for c in working_profile.candidates if c not in curr_cands])
    rcounts = working_profile.rcounts
    
    num_cands = len(candidates)
    ranks = range(1, num_cands + 1)
    
    # Get rankings and their counts
    rankings, rcounts = working_profile.rankings_counts
    
    # Track scores at each rank level
    cand_to_num_voters_rank = dict()
    for r in range(1, num_cands + 1):
        cand_to_num_voters_rank[r] = {c: _num_rank_profile_with_ties(rankings, rcounts, c, r)
                                     for c in candidates}
        # Calculate cumulative scores up to current rank
        cand_scores = {c: sum([cand_to_num_voters_rank[_r][c] for _r in range(1, r + 1)])
                      for c in candidates}
        
        # Check if any candidate has a majority
        if any([s >= strict_maj_size for s in cand_scores.values()]):
            max_score = max(cand_scores.values())
            return sorted([c for c in candidates if cand_scores[c] >= max_score])
    
    # If no candidate has majority, return those with highest cumulative score
    max_score = max(cand_scores.values())
    return sorted([c for c in candidates if cand_scores[c] >= max_score])


@vm(name = "Simplified Bucklin",
    input_types = [ElectionTypes.PROFILE])
def simplified_bucklin(profile, curr_cands = None): 
    """If a candidate has a strict majority of first-place votes, then that candidate is the winner. If no such candidate exists, then check the candidates that are ranked first or second.  If a candidate has a strict majority of first- or second-place voters, then that candidate is the winner. If no such winner is found move on to the 3rd, 4th, etc. place votes. 

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.other_methods import simplified_bucklin

        prof = Profile([[1, 0, 2], [0, 2, 1], [0, 1, 2]], [2, 1, 1])

        prof.display()
        simplified_bucklin.display(prof)

    """
    strict_maj_size = profile.strict_maj_size()
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    
    num_cands = candidates
    
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), profile.num_cands)

    rcounts = profile._rcounts
    
    num_cands = len(candidates)
    ranks = range(1, num_cands + 1)
    
    cand_to_num_voters_rank = dict()
    for r in ranks:
        cand_to_num_voters_rank[r] = {c: _num_rank(rankings, rcounts, c, r)
                                      for c in candidates}
        cand_scores = {c:sum([cand_to_num_voters_rank[_r][c] for _r in cand_to_num_voters_rank.keys()]) 
                       for c in candidates}
        if any([s >= strict_maj_size for s in cand_scores.values()]):
            break
            
    return sorted([c for c in candidates if cand_scores[c] >= strict_maj_size])

def simplified_bucklin_with_explanation(profile, curr_cands = None): 
    """Return the Simplified Bucklin winners and the score for each candidate. 
 
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 

        A sorted list of candidates

        A dictionary assigning the score for each candidate. 

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.other_methods import simplified_bucklin_with_explanation

        prof = Profile([[1, 0, 2], [0, 2, 1], [0, 1, 2]], [2, 1, 1])

        prof.display()
        sb_ws, scores = simplified_bucklin_with_explanation(prof)

        print(f"The winners are {sb_ws}")
        print(f"The candidate scores are {scores}")

    """
    strict_maj_size = profile.strict_maj_size()
    
    candidates = profile.candidates if curr_cands is None else curr_cands
    
    num_cands = candidates
    
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), profile.num_cands)

    rcounts = profile._rcounts
    
    num_cands = len(candidates)
    ranks = range(1, num_cands + 1)
    
    cand_to_num_voters_rank = dict()
    for r in ranks:
        cand_to_num_voters_rank[r] = {c: _num_rank(rankings, rcounts, c, r)
                                      for c in candidates}
        cand_scores = {c:sum([cand_to_num_voters_rank[_r][c] for _r in cand_to_num_voters_rank.keys()]) 
                       for c in candidates}
        if any([s >= strict_maj_size for s in cand_scores.values()]):
            break
            
    return sorted([c for c in candidates if cand_scores[c] >= strict_maj_size]), cand_scores

@vm(name = "Weighted Bucklin",
    input_types = [ElectionTypes.PROFILE])
def weighted_bucklin(profile, curr_cands = None, strict_threshold = False, score = lambda num_cands, rank: (num_cands - rank)/ (num_cands - 1) if num_cands > 1 else 1): 
    """The Weighted Bucklin procedure, studied by D. Marc Kilgour, Jean-Charles Grégoire, and Angèle Foley. The k-th Weighted Bucklin score of a candidate c is the sum for j \leq k of the product of score(num_cands,j) and the number of voters who rank c in j-th place. Compute higher-order Weighted Bucklin scores until reaching a k such that some candidate's k-th Weighted Bucklin score is at least half the number of voters (or the strict majority size if strict_threshold = True). Then return the candidates with maximal k-th Weighted Bucklin score. Bucklin is the special case where strict_threshold = True and score = lambda num_cands, rank: 1.
    
    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        strict_threshold: If True, makes the threshold for the Bucklin procedure the strict majority size; otherwise threshold is half the number of voters, following Kilgour et al.
        score (function): A function that accepts two parameters ``num_cands`` (the number of candidates) and ``rank`` (a rank of a candidate) used to calculate the score of a candidate. The default ``score`` function is the normalized version of the classic Borda score vector.

    Returns: 
        A sorted list of candidates

    :Example: 

    .. exec_code:: 

        from pref_voting.profiles import Profile
        from pref_voting.other_methods import weighted_bucklin

        prof = Profile([[1, 0, 2], [0, 2, 1], [0, 1, 2]], [2, 1, 1])

        prof.display()
        weighted_bucklin.display(prof)

    """
    if strict_threshold == True:
        threshold = profile.strict_maj_size()
    else:
        threshold = profile.num_voters / 2

    candidates = profile.candidates if curr_cands is None else curr_cands
    
    num_cands = candidates
    
    rankings = profile._rankings if curr_cands is None else _find_updated_profile(profile._rankings, np.array([c for c in profile.candidates if c not in curr_cands]), profile.num_cands)

    rcounts = profile._rcounts
    
    num_cands = len(candidates)
    ranks = range(1, num_cands + 1)
    
    cand_to_num_voters_rank = dict()
    for r in ranks:
        cand_to_num_voters_rank[r] = {c: _num_rank(rankings, rcounts, c, r)
                                      for c in candidates}
        cand_scores = {c:sum([score(len(candidates), _r) * cand_to_num_voters_rank[_r][c] for _r in cand_to_num_voters_rank.keys()]) 
                       for c in candidates}
        if any([s >= threshold for s in cand_scores.values()]):
            break
    max_score = max(cand_scores.values())

    return sorted([c for c in candidates if cand_scores[c] >= max_score])

def _dodgson_score(profile, cand):
    """
    Return the **Dodgson score** of ``cand`` in ``profile``.

    The Dodgson score of a candidate *c* is the minimum number of
    *adjacent* swaps, summed over *all* ballots, needed to make *c* the
    **Condorcet winner**.

    This is equivalent to the the minimum number of places that $c$ 
    must *move up* in ballots to become a Condorcet winner
    (see Lemma 4.0.5 of John C. McCabe-Dansted, 
    *Approximability and computational feasibility of Dodgson’s rule*. 
    Master’s thesis, University of Auckland, 2006.)

    We formulate this as a *mixed‑integer program* and solve it with
    OR‑Tools/SCIP:

        • Variables x[r, k] = number of voters with ranking r who move 
        *c* upward by *exactly* k positions (0 ≤ k ≤ pos₍r₎(c)).
        • Objective: minimise Σ k · x[r, k] (total swaps).
        • Constraints
            – Partition: for each distinct ranking r, the x[r,·] must
              sum to the observed multiplicity of r.
            – Condorcet: for every rival d ≠ c, after the moves
              *c* must beat d by a strict majority.

    Parameters
    ----------
    profile : pref_voting.profiles.Profile
    cand : int, the candidate whose Dodgson score we compute.

    Returns
    -------
    int, the Dodgson score of *cand*.
    """
    rankings, counts = profile.rankings_counts   # ranking types & their multiplicities
    majority = profile.strict_maj_size()         # ⌊number of voters / 2⌋ + 1

    solver = pywraplp.Solver.CreateSolver("SCIP")
    if solver is None:
        raise EnvironmentError("This OR‑Tools build lacks a MIP solver.")

    # Variables x[r, k]
    #
    # For each ranking type and each possible upward
    # move k (0..pos_c) we create an *integer* variable representing
    # the number of voters with that ranking who move *c* upward by
    # exactly k positions.
    #
    # The partition constraints (one per ranking) ensure that we do not
    # duplicate or delete ballots but only *re‑order* the existing ones.

    x = {}  # maps (r_idx, k) -> IntVar
    for r_idx, (ranking, w_np) in enumerate(zip(rankings, counts)):
        w = int(w_np)                # NumPy scalar → Python int  (SCIP wants int bounds)
        order = list(ranking)
        pos_c = order.index(cand)    # 0‑based position of cand in this ballot

        # create variables for k = 0 … pos_c
        for k in range(pos_c + 1):
            x[(r_idx, k)] = solver.IntVar(0, w, f"x_{r_idx}_{k}")

        # partition constraint  Σ_k x[r,k] = multiplicity of ranking r
        solver.Add(solver.Sum(x[(r_idx, k)] for k in range(pos_c + 1)) == w)

    # Condorcet constraints:  make cand beat every rival d ≠ cand
    # For each rival d we compare:
    #   current_support   – voters already preferring cand over d
    #   contributed_flips – voters whose ballots we move enough for cand
    #                       to pass d (k ≥ distance in that ranking)
    # The sum must reach the strict majority threshold.
    for d in range(profile.num_cands):
        if d == cand:
            continue  # skip self‑comparison

        current_support = profile.support(cand, d)
        flip_terms = []

        # identify voters who can move cand up by k positions to go above d
        for r_idx, ranking in enumerate(rankings):
            order = list(ranking)
            pos_c = order.index(cand)
            pos_d = order.index(d)
            if pos_d < pos_c:                       # cand currently below d
                dist = pos_c - pos_d                # minimum upward steps to pass d
                for k in range(dist, pos_c + 1):    # any k ≥ dist suffices
                    flip_terms.append(x[(r_idx, k)])

        # enforce majority: support_after ≥ majority
        solver.Add(current_support + solver.Sum(flip_terms) >= majority)

    # Objective: minimise total adjacent swaps  
    # Each voter who moves cand up by k positions contributes exactly k swaps.
    solver.Minimize(
        solver.Sum(k * var for (r_idx, k), var in x.items() if k > 0)
    )

    if solver.Solve() != pywraplp.Solver.OPTIMAL:
        raise RuntimeError("SCIP failed to prove optimality.")

    return int(solver.Objective().Value())

@vm(name="Dodgson", 
    input_types = [ElectionTypes.PROFILE])
def dodgson(profile, curr_cands=None, global_score = True):
    """The Dodgson score of a candidate is the minimum number of adjacent swaps in the ballots needed to make them a Condorcet winner. The Dodgson method selects the candidate with the minimum Dodgson score.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the candidates in curr_cands with the best Dodgson score.
        global_score (bool, optional): If True, then the Dodgson score is computed using the entire profile, including candidates not in curr_cands. If False, then the Dodgson score is computed using the profile restricted to the candidates in curr_cands.

    Returns:
        A sorted list of candidates

    """
    if curr_cands is None:
        curr_cands = profile.candidates

    if not global_score:
        profile.remove_candidates([c for c in profile.candidates if c not in curr_cands])

    scores = {c: _dodgson_score(profile, c) for c in curr_cands}

    best   = min(scores.values())
    return sorted([c for c, s in scores.items() if s == best])

@vm(name = "Bracket Voting",
    input_types = [ElectionTypes.PROFILE])
def bracket_voting(profile, curr_cands = None, seed = None, tie_break = "random"):
    """The candidates with the top four plurality scores are seeded into a bracket: the candidate with the highest plurality score is seeded 1st, the candidate with the second highest plurality score is seeded 2nd, etc. The 1st seed faces the 4th seed in a head-to-head match decided by majority rule, and the 2nd seed faces the 3rd seed in a head-to-head match decided by majority rule. The winners of these two matches face each other in a final head-to-head match decided by majority rule. The winner of the final is the winner of the election.

    .. note::
        A version of bracket voting as proposed by Edward B. Foley. By default, this is a probabilistic method that always returns a unique winner, as ties are broken using a random tie breaking ordering of the candidates.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        seed (int, optional): The seed for the random tie breaking ordering of the candidates.   
        tie_break (str, optional): The method used to break ties in the head-to-head matches. If set to "random", then a random tie breaking ordering of the candidates is used. If set to "lexicographic", then the candidates are sorted lexicographically.
    Returns: 
        A sorted list of candidates

    """
    cands = curr_cands if curr_cands else profile.candidates

    if len(cands) == 2:
        return plurality(profile, curr_cands = curr_cands)
    
    if tie_break == "random":
        rng = np.random.default_rng(seed)
        tie_breaking_ordering = cands.copy()
        rng.shuffle(tie_breaking_ordering)

    elif tie_break == "lexicographic":
        tie_breaking_ordering = sorted(cands)

    plurality_scores = profile.plurality_scores(curr_cands = cands)
    descending_plurality_scores = sorted(plurality_scores.values(), reverse=True)
    
    # If there is a tie for max plurality score, the first seed is the candidate with max plurality score who appears first in the tie breaking ordering
    potential_first_seeds = [c for c in cands if plurality_scores[c] == descending_plurality_scores[0]]
    first_seed = min(potential_first_seeds, key = lambda c: tie_breaking_ordering.index(c)) 

    potential_second_seeds = [c for c in cands if plurality_scores[c] == descending_plurality_scores[1] and c != first_seed]
    second_seed = min(potential_second_seeds, key = lambda c: tie_breaking_ordering.index(c))

    potential_third_seeds = [c for c in cands if plurality_scores[c] == descending_plurality_scores[2] and c not in [first_seed, second_seed]]
    third_seed = min(potential_third_seeds, key = lambda c: tie_breaking_ordering.index(c))

    potential_fourth_seeds = [c for c in cands if plurality_scores[c] == descending_plurality_scores[3] and c not in [first_seed, second_seed, third_seed]] if len(cands) > 3 else []
    fourth_seed = min(potential_fourth_seeds, key = lambda c: tie_breaking_ordering.index(c)) if len(potential_fourth_seeds) > 0 else None

    # Ties in semi-final head-to-head matches are broken in favor of the higher-seeded candidate
    if len(cands) == 3:
        one_four_winner = first_seed
        one_four_winner_seed = 1
    else: 
        one_four_winner = first_seed if profile.margin(first_seed, fourth_seed) >= 0 else fourth_seed
        one_four_winner_seed = 1 if one_four_winner == first_seed else 4

    two_three_winner = second_seed if profile.margin(second_seed, third_seed) >= 0 else third_seed
    two_three_winner_seed = 2 if two_three_winner == second_seed else 3

    if profile.margin(one_four_winner, two_three_winner) > 0:
        winner = one_four_winner

    elif profile.margin(one_four_winner, two_three_winner) < 0:
        winner = two_three_winner

    # Ties in the final head-to-head match are broken in favor of the higher-seeded candidate
    else:
        winner = one_four_winner if one_four_winner_seed < two_three_winner_seed else two_three_winner

    return [winner]

@vm(name = "Superior Voting",
    input_types = [ElectionTypes.PROFILE])
def superior_voting(profile, curr_cands = None):
    """One candidate is superior to another if more ballots rank the first candidate above the second than vice versa. A candidate earns a point from a ballot if they are ranked first on that ballot or they are superior to the candidate ranked first on that ballot. The candidate with the most points wins.

    .. note::
        Devised by Wesley H. Holliday as a simple Condorcet-compliant method for political elections. Always elects a Condorcet winner if one exists and elects only the Condorcet winner provided the Condorcet winner receives at least one first-place vote. Edward B. Foley suggested the name 'Superior Voting' because the method is based on the idea that if A is superior to B, then A should get B's first-place votes added to their own.

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``

    Returns: 
        A sorted list of candidates

    """
    curr_cands = profile.candidates if curr_cands is None else curr_cands

    # Calculate the points for each candidate
    points = {cand: profile.plurality_scores(curr_cands)[cand] for cand in curr_cands}
    for cand in curr_cands:
        for other_cand in curr_cands:
            if profile.margin(cand, other_cand) > 0:
                points[cand] += profile.plurality_scores(curr_cands)[other_cand]
    
    # Find the candidates with the most points
    max_score = max(points.values())
    winners = [cand for cand in curr_cands if points[cand] == max_score]

    return winners

def bt_mle(pmat, max_iter=100):
    """Lucas Maystre's implementation of MLE for the Bradley-Terry model (https://datascience.stackexchange.com/questions/18828/from-pairwise-comparisons-to-ranking-python). 
    
    Note we change the interpretation of p_{i,j} to be the probability that i is preferred to j, rather than vice versa as in the original implementation.
    """
    n = pmat.shape[0]
    wins = np.sum(pmat, axis=1)
    params = np.ones(n, dtype=float)
    for _ in range(max_iter):
        tiled = np.tile(params, (n, 1))
        combined = 1.0 / (tiled + tiled.T)
        np.fill_diagonal(combined, 0)
        nxt = wins / np.sum(combined, axis=0)
        nxt = nxt / np.mean(nxt)
        if np.linalg.norm(nxt - params, ord=np.inf) < 1e-6:
            return nxt
        params = nxt
    raise RuntimeError('did not converge')

@vm(name = "Bradley-Terry",
    input_types = [ElectionTypes.PROFILE])
def bradley_terry(prof, curr_cands = None, threshold = .00001):
    """The Bradley-Terry model is a probabilistic model for pairwise comparisons. In this model, the probability that a voter prefers candidate i to candidate j is given by p_{i,j} = v_i / (v_i + v_j), where v_i is the strength of candidate i. Given a profile, we take p_{i,j} to be the proportion of voters who prefer candidate i to candidate j. We then estimate the strength of each candidate using maximum likelihood estimation. The winning candidates are those whose estimated strength is within +/- threshold of the maximum strength.

    .. note::
        For profiles of linear ballots, this is equivalent to Borda (see Theorem 3.1 of https://arxiv.org/abs/2312.08358).

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        threshold (float, optional): The threshold for determining the winners. The winners are those whose estimated strength is within +/- threshold of the maximum strength.

    Returns: 
        A sorted list of candidates
    """

    curr_cands = prof.candidates if curr_cands is None else curr_cands
    
    prop_matrix = np.zeros((len(curr_cands), len(curr_cands)))

    for i, c in enumerate(curr_cands):
        for j, d in enumerate(curr_cands):
            if i != j:
                prop_matrix[i][j] = prof.support(c,d) / prof.num_voters

    params = bt_mle(prop_matrix)

    max_value = np.max(params)
    winner_indices = np.where(np.abs(params - max_value) <= threshold)[0]
    winners = [curr_cands[i] for i in winner_indices]

    return sorted(winners)

@swf(name = "Bradley-Terry Ranking")
def bradley_terry_ranking(prof, curr_cands = None, threshold = .00001):
    """The Bradley-Terry model is a probabilistic model for pairwise comparisons. In this model, the probability that a voter prefers candidate i to candidate j is given by p_{i,j} = v_i / (v_i + v_j), where v_i is the strength of candidate i. Given a profile, we take p_{i,j} to be the proportion of voters who prefer candidate i to candidate j. We then estimate the strength of each candidate using maximum likelihood estimation. Finally, the candidates are ranked in decreasing order of their estimated strength (where candidates whose estimated strength is within +/- threshold of each other are considered tied).

    .. note::
        For profiles of linear ballots, this is equivalent to Borda (see Theorem 3.1 of https://arxiv.org/abs/2312.08358).

    Args:
        profile (Profile): An anonymous profile of linear orders on a set of candidates
        curr_cands (List[int], optional): If set, then find the winners for the profile restricted to the candidates in ``curr_cands``
        threshold (float, optional): The threshold for equivalence classes of candidates. 

    Returns: 
        A Ranking object.
    """

    curr_cands = prof.candidates if curr_cands is None else curr_cands
    
    support_matrix = np.zeros((len(curr_cands), len(curr_cands)))

    for i, c in enumerate(curr_cands):
        for j, d in enumerate(curr_cands):
            if i != j:
                support_matrix[i][j] = prof.support(c,d) / prof.num_voters

    params = bt_mle(support_matrix)

    ranking_dict = dict()
    cands_assigned = list()
    curr_ranking = 1

    while len(cands_assigned) < len(curr_cands):

        max_value = np.max([params[curr_cands.index(c)] for c in curr_cands if c not in cands_assigned])

        for c in curr_cands:
            if c not in cands_assigned and np.abs(params[curr_cands.index(c)] - max_value) <= threshold:
                ranking_dict[c] = curr_ranking
                cands_assigned.append(c)

        curr_ranking += 1

    return Ranking(ranking_dict)


from pref_voting.utility_methods import sum_utilitarian_ranking, relative_utilitarian_ranking, maximin_ranking, lexicographic_maximin_ranking, nash_ranking, utilitarian_swfs

from pref_voting.scoring_methods import plurality_ranking, borda_ranking, scoring_swfs

# List of all social welfare functions
social_welfare_functions = utilitarian_swfs + scoring_swfs




"""
File: other_axioms.py
Date: April 29 2025
Authors: Wes Holliday (wesholliday@berkeley.edu) and Eric Pacuit (epacuit@umd.edu)

Other axioms
---------------------
"""

from itertools import repeat
import numpy as np

from pref_voting.axiom import Axiom
from pref_voting.profiles import Profile
from pref_voting.profiles_with_ties import ProfileWithTies
from pref_voting.rankings import Ranking


def _reverse_ranking(ballot, all_cands):
    """
    Reverse a single ballot, treating *unranked* candidates as
    tied for last before the reversal.

    Parameters
    ----------
    ballot : tuple | Ranking
        • tuple  – a strict linear order
        • Ranking – weak order, possibly incomplete
    all_cands : list | set
        The full candidate set of the profile.
    """
    if isinstance(ballot, tuple):                 
        return tuple(reversed(ballot))              

    if isinstance(ballot, Ranking):
        full_rmap = ballot.rmap.copy()
        if full_rmap:
            last_rank = max(full_rmap.values()) + 1
        else:                                        
            last_rank = 1
        for c in all_cands:
            if c not in full_rmap:
                full_rmap[c] = last_rank

        max_rank = max(full_rmap.values())
        rev_rmap = {c: max_rank + 1 - k for c, k in full_rmap.items()}
        return Ranking(rev_rmap)

def _reverse_profile(P):
    """
    Build the reversed profile ``Pᵣ``.
    """
    all_cands = P.candidates
    rev_ballots = [_reverse_ranking(b, all_cands) for b in P.rankings]

    if isinstance(P, Profile):
        return Profile(rev_ballots)

    if isinstance(P, ProfileWithTies):
        P_r = ProfileWithTies(rev_ballots, candidates=all_cands)
        if P.using_extended_strict_preference:
            P_r.use_extended_strict_preference()
        return P_r
    
def _reverse_margin_graph(mg):
    """
    Return the *edge-reversed* margin graph.

    All positive-margin edges (u, v, w) become (v, u, w); weights are preserved.
    """
    rev_edges = [(v, u, w) for (u, v, w) in mg.edges]
    return MarginGraph(mg.candidates[:], rev_edges, cmap=mg.cmap)

def has_reversal_symmetry_violation(edata, vm, verbose=False):
    """
    Returns True iff ``vm`` violates reversal symmetry on *edata*.

    Reversal Symmetry states that if x is a **unique** winner in ``edata``, 
    then x should not be among the winners in the reversal of ``edata``.
    """

    if len(edata.candidates) <= 1:
        return False

    if isinstance(edata, MarginGraph):
        mg = edata
        winners = vm(mg)
        if len(winners) != 1:
            return False

        x = winners[0]
        mg_r = _reverse_margin_graph(mg)
        rev_winners = vm(mg_r)

        if x in rev_winners:
            if verbose:
                print(f"Reversal-symmetry violation for {vm.name} on a MarginGraph")
                print(f"Unique winner {x} also wins after edge reversal.")
                print("\nOriginal margin graph:")
                mg.display()
                print(mg.description())
                vm.display(mg)
                print('\nReversed margin graph:')
                mg_r.display()
                print(mg_r.description())
                vm.display(mg_r)

            return True
        return False

    winners = vm(edata)
    if isinstance(winners, np.ndarray):
        winners = winners.tolist()
    if len(winners) != 1:
        return False

    x = winners[0]
    P_r = _reverse_profile(edata)
    rev_winners = vm(P_r)
    if isinstance(rev_winners, np.ndarray):
        rev_winners = rev_winners.tolist()

    if x in rev_winners:
        if verbose:
            print(f"Reversal-symmetry violation for {vm.name}")
            print(f"Unique winner {x} also wins after reversal.")
            print("\nOriginal profile:")
            edata.display()
            print(edata.description())
            vm.display(edata)
            print("\nReversed profile:")
            P_r.display()
            print(P_r.description())
            vm.display(P_r)
        return True
    return False


def find_all_reversal_symmetry_violations(edata, vm, verbose=False):
    """
    Returns a one-item list [(unique_winner, winners_after_reversal)] describing the violation on *edata*, or [].
    """
    if not has_reversal_symmetry_violation(edata, vm, verbose):
        return []

    winners = vm(edata)
    winners = winners.tolist() if isinstance(winners, np.ndarray) else winners
    if isinstance(edata, MarginGraph):
        rev_winners = vm(_reverse_margin_graph(edata))
    else:
        rev_winners = vm(_reverse_profile(edata))

    rev_winners = rev_winners.tolist() if isinstance(rev_winners, np.ndarray) else rev_winners
    return [(winners, rev_winners)]


reversal_symmetry = Axiom(
    "Reversal Symmetry",
    has_violation=has_reversal_symmetry_violation,
    find_all_violations=find_all_reversal_symmetry_violations,
)